#include "BspNode.h"

#include "Mesh.h"

#include <algorithm>

/**
 * The BspNode constructor takes a vector of triangles and a boolean flag indicating whether to split
 * the triangles, and recursively constructs a binary space partitioning tree.
 * 
 * @param triangles A vector of Triangle objects representing the triangles to be used for constructing
 * the BSP tree.
 * @param split A boolean flag indicating whether or not to split the triangles during the construction
 * of the BSP tree.
 */
BspNode::BspNode( const std::vector<Triangle> &triangles, const bool &split )
    : m_front( nullptr )
    , m_back( nullptr )
{
    /**
     * TODO: Aufgabe 1b
     * 
     * Erzeugen Sie hier einen BSP-Baum, welcher alle Dreiecke `triangles` umfasst. Ob Dreiecke
     * gesplittet werden entscheidet das Flag `split`. Die Vorderseite des Unterbaumes wird in `m_front`
     * gespeichert, die Rückseite in `m_back`. Es bietet sich an, diese Funktion rekursiv zu
     * implementieren, d.h., sollte einer der Unterbäume nicht leer werden (erkennbar an Hand der Anzahl
     * der Dreiecke vor der Ebene und hinter der Ebene), dann wird diese Funktion rekursiv für den
     * nicht-leeren Unterbaum aufgerufen. Beispiel:  `m_front = new BspNode( front, split );`
     * Achten Sie wie gesagt darauf, dass wenn `front` leer ist der Aufruf nicht geschieht! Ansonsten
     * gibt es eine endlose Schleife, welche zu einem Stackoverflow führen wird.
     * Wenn `m_front` den Wert `nullptr` hat ist es ein Indikator dafür, dass der Baum hier nicht
     * weitergeht, somit lassen Sie diesen Wert, welcher auch initial dort steht, erhalten, wenn der
     * Baum dort nicht tiefer geht (analog mit `m_back`).
     *
     * Für einen neuen Knoten müssen Sie ein Dreieck auswählen und dieses in `m_triangle` speichern,
     * und die dazugehörige Ebene in `m_planeNormal` speichern (wir brauchen keinen extra Punkt, da
     * ein beliebiger Punkt von `m_triangle` als Stützpunkt der Ebene genügt). Für die Berechnung der
     * Ebene können Sie die statische Funktion `CreateNormal()` nutzen (diese ist in `BspNode.h`
     * definiert).
     * Nachdem Sie ein Dreieck ausgewählt haben müssen Sie über alle restlichen(!) Dreiecke iterieren
     * und diese ohne Splitting in zwei lokale Listen `front` und `back` sortieren, je nachdem ob diese
     * vor oder hinter der Ebene dieses Knotens liegen (ich definiere vorne in Richtung der Normale die
     * bei CreateNormal erzeugt wurde). Ohne Splitting müssen Sie hier immer entscheiden auf welche Seite
     * das Dreieck gehört, auch wenn es die Ebene schneidet.
     * Sollte Splitting aktiviert sein, dann können Sie versuchen mit Hilfe der Funktion `splitTriangle`
     * das Splitting durchzuführen, die Funtion liefert per Boolean `true`, wenn das Dreieck erfolgreich
     * gesplittet wurde und fügt die neuen Dreieck an die Liste, die übergeben werden an, Sie müssen für
     * diese neuen Dreiecke wieder die Sortierung in die Listen `front` und `back` vornehmen, an Hand der
     * Seite der Ebene auf welcher sie liegen. Sollte das Splitting nicht erforderlich gewesen sein,
     * müssen Sie wie im vorherigen Paragraphen beschrieben vorgehen (d.h. sortieren Sie das ungesplittete
     * Dreieck je nach dem auf welcher Seite der Ebene es liegt in eine der beiden Listen ein).
     */

    // m_triangle = ...; // Weisen Sie dem Knoten ein Dreieck zu
    // m_planeNormal = ...; // Erzeugen Sie eine Ebene in der das Dreieck liegt
    // Iteration über restliche Dreiecke
    // Rekursion
}

/**
 * @brief Auxiliary function to create a normal for a triangle.
 * 
 * @param v0 One of the vertices.
 * @param v1 Another vertex.
 * @param v2 Another vertex.
 * @return glm::vec3 A normal vector.
 */
glm::vec3 CreateNormal( glm::vec3 v0, glm::vec3 v1, glm::vec3 v2 ) {
    /**
     * TODO: Aufgabe 1a
     * 
     * Berechnen Sie die Normale des Dreiecks v0,v1,v2. Die Normale liegt
     * senkrecht auf dem Dreieck und hat die Länge 1.
     */
}

/**
 * The destructor for the BspNode class deletes the front and back child nodes.
 */
BspNode::~BspNode()
{
    /**
     * TODO: Aufgabe 1c
     * 
     * Geben Sie die im Konstrutor allozierten Heapspeicher (das passiert immer dann, wenn Sie `new`
     * benutzt haben) frei, in dem Sie `delete` auf den Pointer aufrufen.
     * Durch die Rekursion beim Aufbau muss der Dekonstruktor auch rekursiv auf die Kinderknoten
     * aufgerufen werden (wenn sie `delete` aufrufen passiert genau das).
     */
}

/**
 * The function gathers and sorts triangles based on the eye position and the plane normal.
 * Recursive definition, this call will call the subtree's gather function to traverse the
 * whole tree.
 * 
 * @param sortedTriangles A reference to a vector of Triangle objects. This vector will be populated
 * with the sorted triangles during the recursive gathering process.
 * @param eyePosition The eyePosition parameter represents the position of the viewer or camera in 3D
 * space. It is used to determine which side of the plane the viewer is on, in order to decide which
 * side of the BSP tree to render first.
 */
void BspNode::gather( std::vector<Triangle> &sortedTriangles, const glm::vec3 &eyePosition )
{
    /**
     * TODO: Aufgabe 2
     *
     * Hier soll nun die vorher erstellte Szene gerendert werden, dazu gibt es eine Kameraposition
     * `eyePosition`, welche entscheiden soll von wo wir anfangen die Dreiecke zu rendern. Diese
     * sollen möglichst weit weg von der Kamera zuerst gerendert werden, so dass Dreiecke die vor
     * anderen liegen zuletzt gerendert werden, damit wird visuell korrekte Transparenz erzeugt.
     * 
     * Wie bei der Erstellung des BSP ist die Traversierung auch rekursiv, d.h. Sie sollten
     * zunächst entscheiden, auf welcher Seite der Ebene des aktuellen Knotens liegt die Kamera,
     * und fangen dann an auf der gegenüberliegenden Seite zu rendern, dann wird das Dreieck von
     * diesem Knoten gerendert, zuletzt die Dreiecke die auf der Seite der Kamera liegen. Die
     * Seiten werden wieder per Rekursion behandelt.
     * Achten Sie darauf, dass falls `m_front` oder `m_back` den Wert `nullptr` haben, dass Sie
     * nicht versuchen diese zu dereferenzieren, z.B. per `*m_back` oder `m_back->function()`.
     * Bei dem Wert `nullptr` hört der Baum auf, somit soll auch die Rekursion aufhören.
     * 
     * In dieser Funktion werden lediglich die Dreiecke in der Liste `sortedTriangles` hinzugefügt.
     * Die Reihenfolge dieser Liste entscheidet jedoch danach in welcher Reihenfolge diese gerendert
     * werden.
     */

    // sortedTriangles.push_back( ... ); // Füge Dreiecke die gerendert werden sollen in die Liste ein
}

/**
 * The function splitTriangle takes a triangle and splits it into two triangles based on a plane
 * defined by a point and a normal vector.
 * 
 * @param tri The triangle that is to be split.
 * @param p A point on the splitting plane. 
 * @param n The normal of the splitting plane.
 * @param newTriangles A vector of triangles that have been created during splitting. This
 * parameter is passed by reference and modified in case splitting actually happens. Any previous
 * triangle is removed in that case!
 * 
 * @return a boolean value that marks whether a split occured, otherwise the no-split type of
 * recursion needs to be done instead.
 */
bool BspNode::splitTriangle( const Triangle& tri, const glm::vec3& p, const glm::vec3& n, std::vector<Triangle>& newTriangles )
{
	const float EPS = 0.01f;

    glm::vec3 a = tri.a.position;
    glm::vec3 b = tri.b.position;
    glm::vec3 c = tri.c.position;

	float d0 = glm::dot( a-p, n );
	float d1 = glm::dot( b-p, n );
	float d2 = glm::dot( c-p, n );

	// sort vertices along signed distance to plane (by little bubble sort)
	if ( d0 > d1 ) {
		std::swap( d0, d1 ); std::swap( a, b );
	}
    if ( d1 > d2 ) {
		std::swap( d1, d2 ); std::swap( b, c );
	}
	if ( d0 > d1 ) {
		std::swap( d0, d1); std::swap( a, b );
	}

	// split tri into triangles
	if ( d0 < -EPS && d2 > EPS ) // we have an intersection
	{
		Triangle newTri;
		if (d1 < -EPS) {
			glm::vec3 d( intersectPlaneLine( p, n, b, c ) );
			glm::vec3 e( intersectPlaneLine( p, n, a, c ) );
            newTriangles.clear();
			newTriangles.push_back( Triangle({Vertex(c),Vertex(d),Vertex(e)}) );
			newTriangles.push_back( Triangle({Vertex(a),Vertex(b),Vertex(d)}) );
			newTriangles.push_back( Triangle({Vertex(a),Vertex(d),Vertex(e)}) );
            return true;
		} else if ( d1 > EPS ) {
			glm::vec3 d( intersectPlaneLine( p, n, b, a ) );
			glm::vec3 e( intersectPlaneLine( p, n, c, a ) );
            newTriangles.clear();
			newTriangles.push_back( Triangle({Vertex(a),Vertex(d),Vertex(e)}) );
			newTriangles.push_back( Triangle({Vertex(c),Vertex(b),Vertex(d)}) );
			newTriangles.push_back( Triangle({Vertex(c),Vertex(d),Vertex(e)}) );
            return true;
		} else {
            // d1 is in the plane
			glm::vec3 e( intersectPlaneLine( p, n, a, c ) );
            newTriangles.clear();
			newTriangles.push_back( Triangle({Vertex(a),Vertex(b),Vertex(e)}) );
			newTriangles.push_back( Triangle({Vertex(c),Vertex(b),Vertex(e)}) );
            return true;
		}
	}
    return false;
}

/**
 * The function calculates the intersection point between a plane and a line segment in 3D space.
 * 
 * @param p The point on the plane.
 * @param n The parameter "n" represents the normal vector of the plane.
 * @param v1 The parameter v1 represents the starting point of the line segment.
 * @param v2 The parameter v2 represents the end point of a line segment.
 * 
 * @return a glm::vec3, which is a 3D vector.
 */
glm::vec3 BspNode::intersectPlaneLine( const glm::vec3 &p, const glm::vec3 &n, const glm::vec3 &v1, const glm::vec3 &v2 )
{
	glm::vec3 dv = v2 - v1;
	float s = (glm::dot(n,p)-glm::dot(n,v1)) / glm::dot(n,dv);
	return v1 + dv*s;
}